@@ -137,9 +137,10 @@ class Hooks:
 
         return name
 
-    def _process_doc(self, thing, data) -> typing.Optional[typing.List[str]]:
+    def _process_doc(
+        self, thing, data, append_prefix=""
+    ) -> typing.Optional[typing.List[str]]:
         doc = ""
-        doc_quoted: typing.Optional[typing.List[str]] = None
 
         if data.doc is not None:
             doc = data.doc
@@ -147,6 +148,17 @@ class Hooks:
             doc = thing["doxygen"]
             doc = sphinxify.process_raw(doc)
 
+        if data.doc_append is not None:
+            doc += f"\n{append_prefix}" + data.doc_append.replace(
+                "\n", f"\n{append_prefix}"
+            )
+
+        return self._quote_doc(doc)
+
+    def _quote_doc(
+        self, doc: typing.Optional[str]
+    ) -> typing.Optional[typing.List[str]]:
+        doc_quoted: typing.Optional[typing.List[str]] = None
         if doc:
             # TODO
             doc = doc.replace("\\", "\\\\").replace('"', '\\"')
@@ -210,7 +222,7 @@ class Hooks:
                 v_data = EnumValue()
             v["x_name"] = self._set_name(name, v_data, strip_prefixes)
             v["data"] = v_data
-            v["x_doc_quoted"] = self._process_doc(v, v_data)
+            v["x_doc_quoted"] = self._process_doc(v, v_data, append_prefix="  ")
 
     def header_hook(self, header, data):
         """Called for each header"""
@@ -244,6 +256,11 @@ class Hooks:
                 qualname = f"::{qualname}"
             tmpl_data_d = tmpl_data.dict()
             tmpl_data_d["x_qualname_"] = qualname.translate(self._qualname_trans)
+            tmpl_data_d["x_doc_set"] = self._quote_doc(tmpl_data.doc)
+            doc_add = tmpl_data.doc_append
+            if doc_add:
+                doc_add = f"\n{doc_add}"
+            tmpl_data_d["x_doc_add"] = self._quote_doc(doc_add)
             self._add_subpackage(tmpl_data_d, tmpl_data)
             templates[k] = tmpl_data_d
 
@@ -779,6 +796,12 @@ class Hooks:
             cls[
                 "x_trampoline_name"
             ] = f"rpygen::Py{cls['x_qualname_']}<typename {cls_qualname}{tmpl}>"
+            cls["x_trampoline_var"] = f"{cls_name}_Trampoline"
+        elif class_data.trampoline_inline_code is not None:
+            raise HookError(
+                f"{cls_key} has trampoline_inline_code specified, but there is no trampoline!"
+            )
+
         cls["x_has_constructor"] = has_constructor
         cls["x_varname"] = "cls_" + cls_name
         cls["x_name"] = self._set_name(cls_name, class_data)
